import readline
import atexit
from .utils.conf import CONF_PATH, MAX_HISTORY
from os import path


class History:
    def __init__(self, filename=path.join(CONF_PATH, 'history')):
        self.filename = filename
        try:
            readline.read_history_file(filename)
            readline.set_history_length(MAX_HISTORY)
        except FileNotFoundError:
            pass
        atexit.register(self.save)

    def save(self):
        try:
            readline.write_history_file(self.filename)
        except Exception:
            pass

    def add(self, line):
        readline.add_history(line)

    def len(self):
        return readline.get_current_history_length()

    def remove_last(self):
        return readline.remove_history_item(self.len() - 1)

    def __del__(self):
        self.save()



from textual.app import App, ComposeResult, Binding
from textual.widgets import Header, Footer, Static, TextArea, Button
from typing import Optional
from ..utils.common import copy_text, execute_cmd, debug_print, extract_code


class PromptInput(Static):
    def __init__(self, prompt):
        self.initial_prompt = prompt
        super().__init__()

    def compose(self) -> ComposeResult:
        yield TextArea(self.initial_prompt, id='prompt_input')


class AnswerOutput(Static):
    def compose(self) -> ComposeResult:
        yield TextArea(id='answer_output')


class CommandOutput(Static):
    def compose(self) -> ComposeResult:
        yield TextArea(id='command_output', read_only=True)


class ButtonDispatch(Static):
    def __init__(self, yank_handler, run_handler):
        super().__init__()
        self.yank_handler = yank_handler
        self.run_handler = run_handler

    def compose(self) -> ComposeResult:
        yield Button(label='Copy', variant='primary', id='copy')
        yield Button(label='Run', variant='error', id='run')

    def on_button_pressed(self, event: Button.Pressed) -> None:
        """Event handler called when a button is pressed."""
        if event.button.id == 'copy':
            self.yank_handler()
        elif event.button.id == 'run':
            self.run_handler()


class ShellGPTApp(App):
    CSS_PATH = 'app.tcss'
    BINDINGS = [
        Binding('ctrl+j', 'infer', 'Infer answer'),
        Binding('ctrl+d', 'toggle_dark', 'Toggle dark mode'),
        Binding('ctrl+y', 'yank', 'Yank code block', priority=True),
        Binding('ctrl+r', 'run', 'Run code block'),
    ]

    def __init__(self, llm, history, initial_prompt):
        self.llm = llm
        self.history = history
        self.has_inflight_req = False
        self.initial_prompt = initial_prompt
        super().__init__()

    def on_mount(self) -> None:
        self.title = f'ShellGPT({self.llm.model})'

    def compose(self) -> ComposeResult:
        """Create child widgets for the app."""
        yield Header()
        yield PromptInput(self.initial_prompt)
        yield AnswerOutput()
        yield ButtonDispatch(
            lambda: self.action_yank(),
            lambda: self.action_run(),
        )
        yield CommandOutput()
        yield Footer()

    def action_toggle_dark(self) -> None:
        """An action to toggle dark mode."""
        self.theme = (
            'textual-dark' if self.theme == 'textual-light' else 'textual-light'
        )

    def action_infer(self) -> None:
        """An action to infer answer."""
        if self.has_inflight_req:
            return

        self.has_inflight_req = True
        try:
            self.infer_inner()
        except Exception as e:
            answer_output = self.query_one('#answer_output')
            answer_output.load_text(f'Error when infer: {e}')
        finally:
            self.has_inflight_req = False

    def get_prompt_input(self) -> Optional[str]:
        prompt_input = self.query_one('#prompt_input')
        prompt = prompt_input.text.strip()
        return None if prompt == '' else prompt

    def get_answer_output(self) -> Optional[str]:
        out = self.query_one('#answer_output')
        text = out.text.strip()
        return None if text == '' else text

    def infer_inner(self) -> None:
        prompt = self.get_prompt_input()
        if prompt is None:
            return

        debug_print(f'infer {prompt}')
        self.history.add(prompt)
        resp = self.llm.chat(prompt)
        buf = ''
        for item in resp:
            buf += item

        script = extract_code(buf)
        if script is not None:
            buf = script

        debug_print(f'infer ret {buf}')
        answer_output = self.query_one('#answer_output')
        answer_output.load_text(buf)

    def action_yank(self) -> None:
        text = self.get_answer_output()
        if text is None:
            return

        copy_text(text)

    def action_run(self) -> None:
        text = self.get_answer_output()
        if text is None:
            return

        cmd = text
        output = execute_cmd(cmd)
        command_output = self.query_one('#command_output')
        command_output.load_text(output)



from shellgpt.utils.common import extract_code, prepare_prompt
from shellgpt.utils.conf import DEFAULT_IMAGE_DIR
from os import path
import unittest


class TestCommon(unittest.TestCase):
    def test_prepare_prompt(self):
        for args, expected in [
            ('hello', ('hello', [])),
            ('https://liujiacai.net/', ('https://liujiacai.net/', [])),
            ('//', ('//', [])),
            ('hello @@test.png', ('hello', [path.join(DEFAULT_IMAGE_DIR, 'test.png')])),
            ('hello /tmp/test.png', ('hello', ['/tmp/test.png'])),
            ('hello/tmp/te\nst.png', ('hello/tmp/te\nst.png', [])),
            ('hello/tmp/test.png', ('hello/tmp/test.png', [])),
            ('hello@@test.png', ('hello@@test.png', [])),
        ]:
            self.assertEqual(prepare_prompt(args), expected)

    def test_extract_code(self):
        for args, expected in [
            ('1+1', None),
            (
                """
            ```
            1+1
            ```
            """,
                '1+1',
            ),
            (
                """
            ```python
            1+1
            ```
            """,
                '1+1',
            ),
        ]:
            self.assertEqual(extract_code(args), expected)



from shellgpt.api.llm import LLM
import unittest

from shellgpt.utils.conf import SYSTEM_CONTENT


class TestLLM(unittest.TestCase):
    def test_make_message_default(self):
        llm = LLM('url', 'key', 'llama3', 'default', 0.8, 10, 2)
        for prompt, expected_msg in [
            ('111', [{'content': '111', 'role': 'user'}]),
            (
                '222',
                [
                    {'content': '111', 'role': 'user'},
                    {'content': '222', 'role': 'user'},
                ],
            ),
            (
                '333',
                [
                    {'content': '222', 'role': 'user'},
                    {'content': '333', 'role': 'user'},
                ],
            ),
        ]:
            actual = llm.make_messages(prompt, True, True)
            self.assertEqual(actual[0], expected_msg)
            self.assertEqual(actual[1], 'llama3')

        # with add_system_message being False
        actual = llm.make_messages('444', True, False)
        self.assertEqual(actual[0], [{'content': '444', 'role': 'user'}])

        # with add_system_message being True again
        actual = llm.make_messages('555', True, True)
        self.assertEqual(
            actual[0],
            [{'content': '333', 'role': 'user'}, {'content': '555', 'role': 'user'}],
        )

    def test_make_message_typo(self):
        llm = LLM('url', 'key', 'llama3', 'typo', 0.8, 10, 2)

        actual = llm.make_messages('hi', True, True)
        self.assertEqual(
            actual[0],
            [
                {'content': SYSTEM_CONTENT.get('typo'), 'role': 'system'},
                {'content': 'hi', 'role': 'user'},
            ],
        )

    def test_make_message_dynamic(self):
        llm = LLM('url', 'key', 'llama3', 'dynamic system body', 0.8, 10, 2)

        actual = llm.make_messages('hi', True, True)
        self.assertEqual(
            actual[0],
            [
                {'content': 'dynamic system body', 'role': 'system'},
                {'content': 'hi', 'role': 'user'},
            ],
        )



import argparse
import sys
import readline
from os import makedirs
from .api.llm import LLM
from .utils.conf import (
    API_KEY,
    MAX_CHAT_MESSAGES,
    MODEL,
    INFER_TIMEOUT,
    API_URL,
    TEMPERATURE,
    SYSTEM_CONTENT,
    CONF_PATH,
    IS_TTY,
)
from .utils.common import (
    is_verbose,
    load_contents_from_config,
    execute_cmd,
    copy_text,
    read_stdin,
    extract_code,
    set_verbose,
    AppMode,
)
from .history import History

__version__ = '0.5.5'


def init_app():
    print(f'Create {CONF_PATH}...')
    makedirs(CONF_PATH, exist_ok=True)


def list_content():
    load_contents_from_config(True)
    for n in SYSTEM_CONTENT:
        print(n)


# List of commands for autocompletion
commands = ['exit', 'clear', 'editor', 'set', 'copy', 'explain', 'run', 'clear']


def completer(text, state):
    options = [cmd for cmd in commands if cmd.startswith(text)]
    if state < len(options):
        return options[state]
    else:
        return None


def repl_setup():
    readline.set_completer(completer)
    # Use Tab for completion
    if sys.platform == 'darwin':
        readline.parse_and_bind('bind ^I rl_complete')
    else:
        readline.parse_and_bind('tab: complete')

    # Enable case-insensitive completion
    readline.set_completer_delims(r' \t\n;')


class ShellGPT(object):
    def __init__(
        self,
        url,
        key,
        model,
        system_content,
        temperature,
        timeout,
        max_messages,
        history,
    ):
        self.is_shell = system_content == 'shell'
        self.last_answer = None
        self.history = history
        self.num_prompt = 0
        self.llm = LLM(
            url, key, model, system_content, temperature, timeout, max_messages
        )

    def tui(self, history, initial_prompt):
        try:
            from .tui.app import ShellGPTApp

            app = ShellGPTApp(self.llm, history, initial_prompt)
            app.run()
        except ImportError:
            print(
                'TUI dependencies are NOT installed. Please install them with: pip install shgpt[tui]'
            )
            sys.exit(1)

    def editor(self):
        print('// Entering editor mode (Ctrl+D to finish, Ctrl+C to cancel)')
        prompt = None
        while True:
            try:
                if prompt is None:
                    prompt = input('')
                else:
                    prompt += input('')
            except KeyboardInterrupt:  # Ctrl+C to cancel
                return None
            except EOFError:  # Ctrl+D to finish
                return prompt

    # return true when prompt is a set command
    def repl_action(self, prompt):
        if 'exit' == prompt:
            self.history.remove_last()
            raise EOFError
        elif prompt in ['c', 'copy']:
            copy_text(self.last_answer)
            return True
        elif prompt == 'clear':
            self.llm.messages.clear()
            copy_text(self.last_answer)
            return True
        elif prompt in ['ed', 'editor']:
            new_prompt = self.editor()
            if new_prompt is not None:
                self.infer(new_prompt)
            return True
        elif prompt in ['clear']:
            self.llm.messages = []
            return True

        if self.is_shell:
            if prompt in ['e', 'explain']:
                self.explain_cmd(self.last_answer)
                return True
            elif prompt in ['r', 'run']:
                print(execute_cmd(self.last_answer, ask=True))
                return True

        # Following parse set command
        if prompt.startswith('set') is False:
            return False

        args = prompt.split(' ')
        if len(args) != 3:
            return False

        sub_cmd = args[1]
        if sub_cmd == 'model':
            self.llm.model = args[2]
            return True
        elif sub_cmd == 'system':
            sc = args[2]
            self.is_shell = sc == 'shell'
            self.llm.system_content = sc
            return True

        return False

    def repl(self, initial_prompt):
        print(
            r"""
__      __   _                    _         ___ _        _ _  ___ ___ _____
\ \    / /__| |__ ___ _ __  ___  | |_ ___  / __| |_  ___| | |/ __| _ \_   _|
 \ \/\/ / -_) / _/ _ \ '  \/ -_) |  _/ _ \ \__ \ ' \/ -_) | | (_ |  _/ | |
  \_/\_/\___|_\__\___/_|_|_\___|  \__\___/ |___/_||_\___|_|_|\___|_|   |_|

Type "exit" or ctrl-d to exit; ctrl-c to stop response; "c" to copy last answer;
     "clear" to reset history messages; "ed" to enter editor mode.
When system content is shell , type "e" to explain, "r" to run last command.
""",
            end='',
        )

        repl_setup()
        try:
            self.repl_inner(initial_prompt)
        except EOFError:
            print(f'\n{self.num_prompt} questions asked this time, bye...')
            sys.exit(0)

    def repl_inner(self, initial_prompt):
        while True:
            try:
                self.infer(initial_prompt)
                prompt = input(f'{self.llm.system_content}@{self.llm.model}> ')
                if IS_TTY and self.repl_action(prompt):
                    self.history.remove_last()
                    continue

                self.infer(prompt)
            except KeyboardInterrupt:
                print()

    def infer(self, prompt):
        if prompt == '':
            return

        self.num_prompt += 1
        self.last_answer = ''
        try:
            for r in self.llm.chat(prompt):
                self.last_answer += r
                if self.is_shell is False:
                    print(r, end='', flush=True)

            if self.is_shell:
                shell = extract_code(self.last_answer)
                if shell is not None:
                    self.last_answer = shell
                print(self.last_answer)
            else:
                print()
        except Exception as e:
            print(f'Error when infer: ${e}')
            if is_verbose():
                raise e

    def explain_cmd(self, cmd):
        self.last_answer = ''
        for r in self.llm.chat(
            f'Explain this command: {cmd}', add_system_message=False
        ):
            self.last_answer += r
            print(r, end='', flush=True)
        print()


def main():
    parser = argparse.ArgumentParser(
        prog='sg',
        description='Chat with LLM in your terminal, be it shell generator, story teller, linux-terminal, etc.',
    )

    parser.add_argument(
        '-r', '--repl', action='store_true', help='enter interactive REPL'
    )
    parser.add_argument('-t', '--tui', action='store_true', help='enter TUI mode')
    parser.add_argument(
        '-S',
        '--shell',
        action='store_true',
        help='system content set to `shell`',
    )
    parser.add_argument(
        '-s',
        '--system',
        default='default',
        help='content for system role (default: %(default)s)',
    )
    parser.add_argument(
        '--timeout',
        type=int,
        help='timeout in seconds for each inference (default: %(default)d)',
        default=INFER_TIMEOUT,
    )
    parser.add_argument(
        '--api-url', default=API_URL, help='base API URL (default: %(default)s)'
    )
    parser.add_argument(
        '--api-key', default=API_KEY, help='API Key (default: %(default)s)'
    )
    parser.add_argument(
        '-m',
        '--model',
        default=MODEL,
        help='model (default: %(default)s)',
    )
    parser.add_argument(
        '-M',
        '--max-messages',
        type=int,
        default=MAX_CHAT_MESSAGES,
        help='max history messages (default: %(default)s)',
    )
    parser.add_argument(
        '--temperature',
        default=TEMPERATURE,
        type=float,
        help='increasing the temperature will make the model answer more creatively. (default: %(default).2f)',
    )
    parser.add_argument(
        '--init', action='store_true', help='create required directories'
    )
    parser.add_argument('--list', action='store_true', help='list known system content')
    parser.add_argument('-v', '--verbose', action='store_true', help='verbose mode')
    parser.add_argument(
        '-V', '--version', action='version', version='%(prog)s ' + __version__
    )
    parser.add_argument('prompt', metavar='<prompt>', nargs='*')
    args = parser.parse_args()
    set_verbose(args.verbose)

    if args.init:
        init_app()
        sys.exit(0)
    elif args.list:
        list_content()
        sys.exit(0)

    sin = read_stdin()
    prompt = ' '.join(args.prompt)
    if sin is not None:
        prompt = f'{sin}\n\n{prompt}'

    if args.repl:
        app_mode = AppMode.REPL
    elif args.tui:
        app_mode = AppMode.TUI
    else:
        app_mode = AppMode.REPL if len(prompt) == 0 else AppMode.Direct

    system_content = args.system
    if args.shell or app_mode == AppMode.TUI:
        system_content = 'shell'

    load_contents_from_config(False)

    history = History()
    sg = ShellGPT(
        args.api_url,
        args.api_key,
        args.model,
        system_content,
        args.temperature,
        args.timeout,
        args.max_messages,
        history,
    )
    if prompt != '':
        history.add(prompt)

    if app_mode == AppMode.Direct:
        sg.infer(prompt)
    elif app_mode == AppMode.TUI:
        sg.tui(history, prompt)
    else:
        sg.repl(prompt)



from os import path, environ, pathsep
import sys
import platform

# Configuration
CONF_PATH = path.expanduser(environ.get('SHELLGPT_CONF_DIR', '~/.shellgpt'))
API_URL = environ.get('SHELLGPT_API_URL', 'http://127.0.0.1:11434')
API_KEY = environ.get('SHELLGPT_API_KEY', '')
MODEL = environ.get('SHELLGPT_MODEL', 'llama3')
IMAGE_MODEL = environ.get('SHELLGPT_IMAGE_MODEL', 'llava')
TEMPERATURE = float(environ.get('SHELLGPT_TEMPERATURE', '0.8'))
DEFAULT_IMAGE_DIR = path.expanduser(environ.get('SHELLGPT_IMAGE_DIR', '~/Pictures'))
INFER_TIMEOUT = int(environ.get('SHELLGPT_INFER_TIMEOUT', '15'))  # seconds
MAX_HISTORY = int(environ.get('SHELLGPT_MAX_HISTORY', '1000'))
MAX_CHAT_MESSAGES = int(environ.get('SHELLGPT_MAX_CHAT_MESSAGES', '5'))

# Auto determined configs
OS_NAME = platform.system()
IS_TTY = sys.stdin.isatty()


def get_shell_type():
    if OS_NAME in ('Windows', 'nt'):
        is_powershell = len(environ.get('PSModulePath', '').split(pathsep)) >= 3
        return 'powershell.exe' if is_powershell else 'cmd.exe'
    return path.basename(environ.get('SHELL', '/bin/sh'))


SHELL = get_shell_type()


# Built-in system content for different user cases.
SYSTEM_CONTENT = {
    'default': None,
    'code': """
Provide only code as output without any description.
Provide only code in plain text format without Markdown formatting.
Do not include symbols such as ``` or ```python.
If there is a lack of details, provide most logical solution.
You are not allowed to ask for more details.
For example if the prompt is "Hello world Python", you should return "print('Hello world')".
    """,
    'shell': """
    You are a shell script assistant on {os_name} running {shell},
    Output the best matching shell commands without any other information, or any quotes.
    Make sure it's valid shell command.
    """.format(os_name=OS_NAME, shell=SHELL),
    'commit': "You are now a git commit message writer. I'll give you a list of changes, and you'll reply with a commit message that summarizes these changes in a clear and concise way, keeping the original formatting.",
    'typo': """
    You are now an article correction assistant. You need to find out the input text in the spelling errors, incoherent places, can only return the corrected text, without any explanation. The output keeps the original format and language output, don't modify the style, and keey the code blocks unchanged.
    """,
    'slug': 'You are now a slug generator. I will give you some sentences, and you will reply with a slug version of those sentences. A slug is a URL-friendly version of a title, where spaces are replaced with hyphens, and all characters are lowercased. Do not include any special characters, and keep the output in English.',
}



import requests


# https://blog.apify.com/python-requests-timeout/
class TimeoutSession(requests.Session):
    def __init__(self, timeout=10):
        self.default_timeout = timeout
        super().__init__()

    def request(self, method, url, **kwargs):
        kwargs.setdefault('timeout', self.default_timeout)
        return super().request(method, url, **kwargs)



from enum import Enum
from typing import Optional
from datetime import datetime
import re
import base64
import os
import subprocess
import sys
import pyperclip

from shellgpt.utils.conf import DEFAULT_IMAGE_DIR, IS_TTY, SYSTEM_CONTENT, CONF_PATH

IS_VERBOSE = False


class AppMode(Enum):
    Direct = (1,)
    TUI = (2,)
    REPL = (3,)


def set_verbose(v):
    global IS_VERBOSE
    IS_VERBOSE = v


def is_verbose():
    return IS_VERBOSE


def debug_print(msg):
    if IS_VERBOSE:
        print(msg)


def extract_code(text: str) -> Optional[str]:
    code = re.findall('```(?:.*?)\n(.*?)```', text, re.DOTALL)
    if len(code) == 0:
        return None
    else:
        return code[0].strip()


def now_ms():
    return int(datetime.now().timestamp() * 1000)


def read_stdin():
    if IS_TTY:
        return None
    else:
        buf = ''
        for line in sys.stdin:
            buf += line

        return buf


def copy_text(text):
    if text is None:
        print('No text to copy!')
    else:
        pyperclip.copy(text)


def execute_cmd(cmd, ask=False):
    if 'y' == input(f'Type y to run: `{cmd}`> '):
        return subprocess.getoutput(cmd)


def base64_image(image_path: str) -> str:
    with open(image_path, 'rb') as image_file:
        return base64.b64encode(image_file.read()).decode('utf-8')


# https://www.debuggex.com/r/6b2cfvu8bb_stYGu
FILE_PATH_RE = re.compile(r' (\/|@@)(.*?)\.(jpg|png)(?:\s|$)')


def extract_paths(txt):
    r = re.findall(FILE_PATH_RE, txt)
    return r


def gen_path(prefix, left, suffix):
    name = left + '.' + suffix
    if prefix == '/':
        return prefix + name
    else:
        return os.path.join(DEFAULT_IMAGE_DIR, name)


def prepare_prompt(raw):
    imgs = [
        gen_path(prefix, path, suffix) for (prefix, path, suffix) in extract_paths(raw)
    ]
    after = raw if len(imgs) == 0 else re.sub(FILE_PATH_RE, '', raw)
    return after, imgs


def load_contents_from_toml():
    conf_file = os.path.join(CONF_PATH, 'prompts.toml')
    if not os.path.exists(conf_file):
        debug_print(f'Configuration file not found: {conf_file}')
        return

    # tomllib is available since Python 3.11+
    import tomllib

    with open(conf_file, 'rb') as f:
        conf = tomllib.load(f)
        global SYSTEM_CONTENT
        SYSTEM_CONTENT.update(conf)


def load_contents_from_config(throw_ex=False):
    try:
        load_contents_from_toml()
    except Exception as e:
        debug_print(f'Error when load contents: ${e}')
        if throw_ex:
            raise e



import json

from ..utils.http import TimeoutSession
from ..utils.common import base64_image, debug_print, prepare_prompt
from ..utils.conf import IMAGE_MODEL, SYSTEM_CONTENT
from urllib.parse import urljoin


def get_system_content(sc):
    return SYSTEM_CONTENT.get(sc, sc)


class LLM(object):
    def __init__(
        self, base_url, key, model, system_content, temperature, timeout, max_messages
    ):
        session = TimeoutSession(timeout=timeout)
        if key is not None and key != '':
            session.headers = {'Authorization': f'Bearer {key}'}
            self.use_openai = True
        else:
            self.use_openai = False
        self.base_url = base_url
        self.model = model
        self.http_session = session
        self.system_content = system_content
        self.temperature = temperature
        self.max_messages = max_messages
        self.messages = []

    def chat(self, prompt, stream=True, add_system_message=True):
        return (
            self.chat_openai(prompt, stream, add_system_message)
            if self.use_openai
            else self.chat_ollama(prompt, stream, add_system_message)
        )

    def chat_openai(self, prompt, stream, add_system_message):
        url = urljoin(self.base_url, '/v1/chat/completions')
        messages, model = self.make_messages(
            prompt,
            False,
            add_system_message,
        )
        debug_print(
            f'chat: {prompt} to {url} with model {self.model} system_content {self.system_content} and stream {stream}, messages: \n{messages}'
        )
        payload = {
            'messages': messages,
            'model': model,
            'stream': stream,
            'temperature': self.temperature,
        }
        r = self.http_session.post(url, json=payload, stream=stream)
        if r.status_code != 200:
            raise Exception('Error: ' + r.text)

        answer = ''
        current = b''
        # https://github.com/openai/openai-python#streaming-responses
        # The response is SSE, so we need to parse the response line by line.
        for item in r.iter_content(chunk_size=None):
            # debug_print(f'\nitem: {item}\ncurrent: {current}')
            for msg in item.split(b'\n\n'):
                msg = msg.removeprefix(b'data: ')
                if len(msg) == 0:
                    continue

                current += msg

                # when current end with '}', it maybe the end of the message
                if current[-1] == 125:
                    # msg is a complete JSON message
                    # `data:` may appear in the middle of the message, so we need to remove it again.
                    msg = current.removeprefix(b'data: ')
                    current = b''
                else:
                    continue

                s = msg.decode('utf-8')
                if s == '[DONE]':
                    self.messages.append({'role': 'assistant', 'content': answer})
                    return
                else:
                    try:
                        resp = json.loads(s)
                        for item in resp['choices']:
                            if 'content' not in item['delta']:
                                continue

                            msg = item['delta']['content']
                            answer += msg
                            yield msg
                    except json.JSONDecodeError:
                        # this means the message is not a JSON message, so we need to continue searching next }.
                        current = msg
                        continue

    def make_messages(self, prompt, support_image, add_system_message):
        model = self.model
        if add_system_message is False:
            return [{'role': 'user', 'content': prompt}], model

        after, imgs = prepare_prompt(prompt) if support_image else (prompt, [])
        if len(imgs) > 0:
            imgs = [base64_image(img) for img in imgs]
            self.messages.append({'role': 'user', 'content': after, 'images': imgs})
            model = IMAGE_MODEL
        else:
            self.messages.append({'role': 'user', 'content': prompt})

        if len(self.messages) > self.max_messages:
            self.messages = self.messages[-self.max_messages :]

        msgs = (
            []
            if self.system_content == 'default'
            else [
                {'role': 'system', 'content': get_system_content(self.system_content)}
            ]
        )
        for m in self.messages:
            msgs.append(m)

        return msgs, model

    # https://github.com/ollama/ollama/blob/main/docs/api.md#generate-a-completion
    def chat_ollama(self, prompt, stream, add_system_message):
        model = self.model
        url = urljoin(self.base_url, '/api/chat')
        messages, model = self.make_messages(prompt, True, add_system_message)
        debug_print(
            f'chat: {prompt} to {url} with model {self.model} system_content {self.system_content} and stream {stream}, messages: \n{messages}'
        )

        payload = {
            'messages': messages,
            'model': model,
            'stream': stream,
            'options': {'temperature': self.temperature},
        }

        r = self.http_session.post(url, json=payload, stream=stream)
        if r.status_code != 200:
            raise Exception('Error: ' + r.text)

        answer = ''
        for item in r.iter_content(chunk_size=None):
            resp = json.loads(item)
            if resp['done']:
                self.messages.append({'role': 'assistant', 'content': answer})
            else:
                content = resp['message']['content']
                answer += content
                yield content
